###########################################################################################################################
# SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA SIA #
###########################################################################################################################
"""
Realiza o download de arquivos principais de dados do SIA (PAXXaamm = Procedimentos Ambulatoriais, XXX)
constante do endereço ftp do Datasus em formato "dbc" como um objeto pandas DataFrame e o salva no
formato "parquet". Caso o arquivo de dados já conste da pasta criada automaticamente no módulo folder
é então realizada a leitura desse arquivo que está no formato "parquet".

Também realiza o download de arquivos auxiliares em formato "dbf" e "cnv".
"""

import os

import numpy as np
import pandas as pd
from ftplib import FTP
from zipfile import ZipFile
from dbfread import DBF

from transform.extract.folder import CACHEPATH

if os.name == 'nt':
    from transform.extract.read_windows import read_dbc, read_cnv
elif os.name == 'posix':
    from transform.extract.read_unix_c import read_dbc, read_cnv


def download_SIAXXaamm(base: str, state: str, year: str, month: str, cache: bool=True):
    """
    Realiza o download de um arquivo principal de dados do SIA em formato "dbc" se já não
    existente no formato "parquet" no diretório CACHEPATH e o lê como um objeto pandas DataFrame

    Parâmetros
    ----------
    base: objeto str
        String de tamanho 2 do nome de uma sub-base de dados do SIA
    state: objeto str
        String de tamanho 2 da sigla de um Estado da RFB
    year: objeto str
        String de tamanho 2 do ano
    month: objeto str
        String de tamanho 2 do mês
    cache: objeto bool
        Boolean se o arquivo "dbc" baixado deve ser salvo no formato "parquet"

    Retorno
    -------
    df: objeto pandas DataFrame
        Dataframe que contém os dados de um arquivo principal de dados originalmente em formato "dbc"
    """

    state = state.upper()
    fname = f'{base}{state}{year}{month}.dbc'
    cachefile = os.path.join(CACHEPATH, 'SIA_' + fname.split('.')[0] + '_.parquet')

    if os.path.exists(cachefile):
        df = pd.read_parquet(cachefile)
        return df
    else:
        ftp = FTP('ftp.datasus.gov.br')
        ftp.login()
        ftp.cwd('/dissemin/publicos/SIASUS/200801_/Dados/')
        try:
            ftp.retrbinary(f'RETR {fname}', open(fname, 'wb').write)
        except:
            try:
                ftp.retrbinary(f'RETR {fname.upper()}', open(fname, 'wb').write)
            except:
                raise Exception(f'Could not access {fname}.')
        df = read_dbc(fname, 'iso-8859-1')
        ftp.close()
        if cache:
            df.to_parquet(cachefile)
        return df


def download_table_dbf(file_name):
    """
    Realiza o download de um arquivo auxiliar de dados do SIA em formato "dbf" ou de uma pasta
    "zip" que o contém (se a pasta "zip" já não foi baixada), em seguida o lê como um objeto pandas
    DataFrame e por fim o elimina

    Parâmetros
    ----------
    file_name: objeto str
        String do nome do arquivo "dbf"

    Retorno
    -------
    df: objeto pandas DataFrame
        Dataframe que contém os dados de um arquivo auxiliar de dados originalmente em formato "dbf"
    """

    if ((file_name == 'CADMUN') or (file_name == 'TABUF')):
        fname = file_name + '.DBF'
        ftp = FTP('ftp.datasus.gov.br')
        ftp.login()
        ftp.cwd('/dissemin/publicos/SIM/CID10/TABELAS/')
        ftp.retrbinary(f'RETR {fname}', open(fname, 'wb').write)

    elif file_name == 'rl_municip_regsaud':
        folder = 'base_territorial.zip'
        ftp = FTP('ftp.datasus.gov.br')
        ftp.login()
        ftp.cwd('/territorio/tabelas/')
        ftp.retrbinary(f'RETR {folder}', open(folder, 'wb').write)
        zip = ZipFile(folder, 'r')
        fname = file_name + '.dbf'
        zip.extract(fname)

    else:
        folder = 'TAB_SIA.zip'
        if not os.path.isfile(folder):
            ftp = FTP('ftp.datasus.gov.br')
            ftp.login()
            ftp.cwd('/dissemin/publicos/SIASUS/200801_/Auxiliar/')
            ftp.retrbinary(f'RETR {folder}', open(folder, 'wb').write)
        zip = ZipFile(folder, 'r')
        try:
            fname = file_name + '.DBF'
            zip.extract('TAB_DBF/' + fname)
        except:
            try:
                fname = file_name + '.dbf'
                zip.extract('TAB_DBF/' + fname)
            except:
                raise Exception(f'Could not access {file_name}.')

    if ((file_name == 'CADMUN')
        or (file_name == 'TABUF')
        or (file_name == 'rl_municip_regsaud')):
        dbf = DBF(fname)

    else:
        dbf = DBF('TAB_DBF/' + fname, encoding='iso-8859-1')

    df = pd.DataFrame(iter(dbf))

    if ((file_name == 'CADMUN')
        or (file_name == 'TABUF')
        or (file_name == 'rl_municip_regsaud')):
        os.unlink(fname)

    else:
        os.unlink('TAB_DBF/' + fname)

    return df


def download_table_cnv(file_name):
    """
    Realiza o download de um arquivo auxiliar de dados do SIA em formato "cnv" ou de uma pasta
    "zip" que o contém (se a pasta "zip" já não foi baixada), em seguida o lê como um objeto pandas
    DataFrame

    Parâmetros
    ----------
    file_name: objeto str
        String do nome do arquivo "cnv"

    Retorno
    -------
    df: objeto pandas DataFrame
        Dataframe que contém os dados de um arquivo auxiliar de dados originalmente em formato "cnv"
    """

    if file_name == 'natjur':
        folder = 'TAB_SIH.zip'
        place = '/dissemin/publicos/SIHSUS/200801_/Auxiliar/'

    else:
        folder = 'TAB_SIA.zip'
        place = '/dissemin/publicos/SIASUS/200801_/Auxiliar/'

    if not os.path.isfile(folder):
        ftp = FTP('ftp.datasus.gov.br')
        ftp.login()
        ftp.cwd(place)
        ftp.retrbinary(f'RETR {folder}', open(folder, 'wb').write)
    zip = ZipFile(folder, 'r')
    try:
        fname = file_name + '.CNV'
        zip.extract('CNV/' + fname)
    except:
        try:
            fname = file_name + '.cnv'
            zip.extract('CNV/' + fname)
        except:
            raise Exception(f'Could not access {file_name}.')

    os.rename('CNV/' + fname, 'SIA_' + fname)

    df = read_cnv('SIA_' + fname)

    return df



###########################################################################################################################################################################
#  CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES CNES   #
###########################################################################################################################################################################

import psycopg2


"""
Cria o schema do banco de dados do CNES para atender ao formato do SGBD postgreSQL.
Na verdade o schema abaixo deve atender ao formato dos SGBD Oracle e SQL Server com no máximo pequenas modificações.

"""


# Função de criação das tabelas do banco de dados cnes
def create_tables(connection_data, child_db):
    conn = psycopg2.connect(dbname=connection_data[0],
                            host=connection_data[1],
                            port=connection_data[2],
                            user=connection_data[3],
                            password=connection_data[4])

    cursor = conn.cursor()

    cursor.execute(f'''/*Tabelas comuns a mais de um sub banco de dados do CNES*/

                       /*Tabela dos estabelecimentos de saúde*/
                       CREATE TABLE IF NOT EXISTS {child_db}.cnes("ID"                      VARCHAR(7),
                                                                  "DESCESTAB"               VARCHAR(66),
                                                                  "RSOC_MAN"                VARCHAR(66),
                                                                  "CPF_CNPJ"                VARCHAR(14),
                                                                  "EXCLUIDO"                NUMERIC,
                                                                  "DATAINCL"                DATE,
                                                                  "DATAEXCL"                DATE);

                       /*Tabela dos Estados da RFB*/
                       CREATE TABLE IF NOT EXISTS {child_db}.ufcod("ID"                     VARCHAR(2),
                                                                   "ESTADO"                 VARCHAR(66),
                                                                   "SIGLA_UF"               VARCHAR(66));

                       /*Tabela de regiões de saúde IBGE*/
                       CREATE TABLE IF NOT EXISTS {child_db}.rsaude("ID"                    VARCHAR(5),
                                                                    "REGIAO"                VARCHAR(66));

                       /*Tabela dos municípios de localização de estabelecimentos*/
                       CREATE TABLE IF NOT EXISTS {child_db}.codufmun("ID"                  VARCHAR(6),
                                                                      "MUNNOME"	            VARCHAR(66),
                                                                      "MUNNOMEX"            VARCHAR(66),
                                                                      "MUNCODDV"            VARCHAR(7),
                                                                      "OBSERV"              VARCHAR(66),
                                                                      "SITUACAO"            VARCHAR(66),
                                                                      "MUNSINP"	            VARCHAR(66),
                                                                      "MUNSIAFI"            VARCHAR(66),
                                                                      "UFCOD_ID"            VARCHAR(2),
                                                                      "AMAZONIA"            VARCHAR(66),
                                                                      "FRONTEIRA"           VARCHAR(66),
                                                                      "CAPITAL"             VARCHAR(66),
                                                                      "RSAUDE_ID"           VARCHAR(5),
                                                                      "LATITUDE"            FLOAT,
                                                                      "LONGITUDE"           FLOAT,
                                                                      "ALTITUDE"            FLOAT,
                                                                      "AREA"                FLOAT,
                                                                      "ANOINST"             VARCHAR(66),
                                                                      "ANOEXT"              VARCHAR(66),
                                                                      "SUCESSOR"            VARCHAR(66));

                       /*Tabela se o estabelecimento é pessoa física ou pessoa jurídica*/
                       CREATE TABLE IF NOT EXISTS {child_db}.pfpj("ID"                      VARCHAR(2),
                                                                  "PESSOA"                  VARCHAR(66));

                       /*Tabela do grau de independência do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.nivdep("ID"                    VARCHAR(2),
                                                                    "TIPO"                  VARCHAR(66));

                       /*Tabela dos tipos de gestão*/
                       CREATE TABLE IF NOT EXISTS {child_db}.tpgestao("ID"                  VARCHAR(2),
                                                                      "GESTAO"              VARCHAR(66));

                       /*Tabela dos tipos de administração*/
                       CREATE TABLE IF NOT EXISTS {child_db}.esferaa("ID"                   VARCHAR(2),
                                                                     "ADMINISTRACAO"        VARCHAR(66));

                       /*Tabela dos tipos de retenção de tributos do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.retencao("ID"                  VARCHAR(2),
                                                                      "RETENCAO"            VARCHAR(66));

                       /*Tabela dos tipos de atividade de ensino, se houver*/
                       CREATE TABLE IF NOT EXISTS {child_db}.atividad("ID"                  VARCHAR(2),
                                                                      "ATIVIDADE"           VARCHAR(66));

                       /*Tabela da natureza do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.natureza("ID"                  VARCHAR(2),
                                                                      "NATUREZA"            VARCHAR(66));

                       /*Tabela dos tipos de fluxo de clientela*/
                       CREATE TABLE IF NOT EXISTS {child_db}.clientel("ID"                  VARCHAR(2),
                                                                      "CLIENTELA"           VARCHAR(66));

                       /*Tabela dos tipos de estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.tpunid("ID"                    VARCHAR(2),
                                                                    "TIPO"                  VARCHAR(66));

                       /*Tabela dos turnos de funcionamento do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.turnoat("ID"                   VARCHAR(2),
                                                                     "TURNO"                VARCHAR(66));

                       /*Tabela dos níveis de atendimento do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.nivhier("ID"                   VARCHAR(2),
                                                                     "NIVEL"                VARCHAR(66));

                       /*Tabela dos tipos de prestador dos serviços hospitalares*/
                       CREATE TABLE IF NOT EXISTS {child_db}.tpprest("ID"                   VARCHAR(2),
                                                                     "PRESTADOR"            VARCHAR(66));

                       /*Tabela das naturezas jurídicas de estabelecimentos*/
                       CREATE TABLE IF NOT EXISTS {child_db}.natjur("ID"                    VARCHAR(4),
                                                                    "NATUREZA"              VARCHAR(100));

                       /*Tabela de Arquivos*/
                       CREATE TABLE IF NOT EXISTS {child_db}.arquivos("NOME"                VARCHAR(15),
                                                                      "DIRETORIO"           VARCHAR(66),
                                                                      "DATA_INSERCAO_FTP"   DATE,
                                                                      "DATA_HORA_CARGA"     TIMESTAMP,
                                                                      "QTD_REGISTROS"       INTEGER);

                       /*Tabelas únicas do sub bancos de dados cnes_st*/

                       /*Tabela Estabelecimentos (tabela principal do sub banco de dados cnes_st)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.stbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_ST"                   VARCHAR(2),
                                                                  "ANO_ST"                  INTEGER,
                                                                  "MES_ST"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "COD_CEP"                 VARCHAR(8),
                                                                  "CPF_CNPJ"                VARCHAR(14),
                                                                  "PFPJ_ID"                 VARCHAR(2),
                                                                  "NIVDEP_ID"               VARCHAR(2),
                                                                  "CNPJ_MAN"                VARCHAR(14),
                                                                  "CODIR_ID"                VARCHAR(2),
                                                                  "VINC_SUS"                NUMERIC,
                                                                  "TPGESTAO_ID"             VARCHAR(2),
                                                                  "ESFERAA_ID"              VARCHAR(2),
                                                                  "RETENCAO_ID"             VARCHAR(2),
                                                                  "ATIVIDAD_ID"             VARCHAR(2),
                                                                  "NATUREZA_ID"             VARCHAR(2),
                                                                  "CLIENTEL_ID"             VARCHAR(2),
                                                                  "TPUNID_ID"               VARCHAR(2),
                                                                  "TURNOAT_ID"              VARCHAR(2),
                                                                  "NIVHIER_ID"              VARCHAR(2),
                                                                  "TPPREST_ID"              VARCHAR(2),
                                                                  "CO_BANCO"                VARCHAR(3),
                                                                  "CO_AGENC"                VARCHAR(5),
                                                                  "C_CORREN"                VARCHAR(14),
                                                                  "ALVARA"                  VARCHAR(25),
                                                                  "DT_EXPED"                DATE,
                                                                  "ORGEXPED_ID"             VARCHAR(2),
                                                                  "AV_ACRED"                NUMERIC,
                                                                  "CLASAVAL_ID"             VARCHAR(2),
                                                                  "DT_ACRED"                DATE,
                                                                  "AV_PNASS"                NUMERIC,
                                                                  "DT_PNASS"                DATE,
                                                                  "GESPRG1E"                NUMERIC,
                                                                  "GESPRG1M"                NUMERIC,
                                                                  "GESPRG2E"                NUMERIC,
                                                                  "GESPRG2M"                NUMERIC,
                                                                  "GESPRG4E"                NUMERIC,
                                                                  "GESPRG4M"                NUMERIC,
                                                                  "NIVATE_A"                NUMERIC,
                                                                  "GESPRG3E"                NUMERIC,
                                                                  "GESPRG3M"                NUMERIC,
                                                                  "GESPRG5E"                NUMERIC,
                                                                  "GESPRG5M"                NUMERIC,
                                                                  "GESPRG6E"                NUMERIC,
                                                                  "GESPRG6M"                NUMERIC,
                                                                  "NIVATE_H"                NUMERIC,
                                                                  "QTLEITP1"                FLOAT,
                                                                  "QTLEITP2"                FLOAT,
                                                                  "QTLEITP3"                FLOAT,
                                                                  "LEITHOSP"                NUMERIC,
                                                                  "QTINST01"                FLOAT,
                                                                  "QTINST02"                FLOAT,
                                                                  "QTINST03"                FLOAT,
                                                                  "QTINST04"                FLOAT,
                                                                  "QTINST05"                FLOAT,
                                                                  "QTINST06"                FLOAT,
                                                                  "QTINST07"                FLOAT,
                                                                  "QTINST08"                FLOAT,
                                                                  "QTINST09"                FLOAT,
                                                                  "QTINST10"                FLOAT,
                                                                  "QTINST11"                FLOAT,
                                                                  "QTINST12"                FLOAT,
                                                                  "QTINST13"                FLOAT,
                                                                  "QTINST14"                FLOAT,
                                                                  "URGEMERG"                NUMERIC,
                                                                  "QTINST15"                FLOAT,
                                                                  "QTINST16"                FLOAT,
                                                                  "QTINST17"                FLOAT,
                                                                  "QTINST18"                FLOAT,
                                                                  "QTINST19"                FLOAT,
                                                                  "QTINST20"                FLOAT,
                                                                  "QTINST21"                FLOAT,
                                                                  "QTINST22"                FLOAT,
                                                                  "QTINST23"                FLOAT,
                                                                  "QTINST24"                FLOAT,
                                                                  "QTINST25"                FLOAT,
                                                                  "QTINST26"                FLOAT,
                                                                  "QTINST27"                FLOAT,
                                                                  "QTINST28"                FLOAT,
                                                                  "QTINST29"                FLOAT,
                                                                  "QTINST30"                FLOAT,
                                                                  "ATENDAMB"                NUMERIC,
                                                                  "QTINST31"                FLOAT,
                                                                  "QTINST32"                FLOAT,
                                                                  "QTINST33"                FLOAT,
                                                                  "CENTRCIR"                NUMERIC,
                                                                  "QTINST34"                FLOAT,
                                                                  "QTINST35"                FLOAT,
                                                                  "QTINST36"                FLOAT,
                                                                  "QTINST37"                FLOAT,
                                                                  "CENTROBS"                NUMERIC,
                                                                  "QTLEIT05"                FLOAT,
                                                                  "QTLEIT06"                FLOAT,
                                                                  "QTLEIT07"                FLOAT,
                                                                  "QTLEIT09"                FLOAT,
                                                                  "QTLEIT19"                FLOAT,
                                                                  "QTLEIT20"                FLOAT,
                                                                  "QTLEIT21"                FLOAT,
                                                                  "QTLEIT22"                FLOAT,
                                                                  "QTLEIT23"                FLOAT,
                                                                  "QTLEIT32"                FLOAT,
                                                                  "QTLEIT34"                FLOAT,
                                                                  "QTLEIT38"                FLOAT,
                                                                  "QTLEIT39"                FLOAT,
                                                                  "QTLEIT40"                FLOAT,
                                                                  "CENTRNEO"                NUMERIC,
                                                                  "ATENDHOS"                NUMERIC,
                                                                  "SERAP01P"                NUMERIC,
                                                                  "SERAP01T"                NUMERIC,
                                                                  "SERAP02P"                NUMERIC,
                                                                  "SERAP02T"                NUMERIC,
                                                                  "SERAP03P"                NUMERIC,
                                                                  "SERAP03T"                NUMERIC,
                                                                  "SERAP04P"                NUMERIC,
                                                                  "SERAP04T"                NUMERIC,
                                                                  "SERAP05P"                NUMERIC,
                                                                  "SERAP05T"                NUMERIC,
                                                                  "SERAP06P"                NUMERIC,
                                                                  "SERAP06T"                NUMERIC,
                                                                  "SERAP07P"                NUMERIC,
                                                                  "SERAP07T"                NUMERIC,
                                                                  "SERAP08P"                NUMERIC,
                                                                  "SERAP08T"                NUMERIC,
                                                                  "SERAP09P"                NUMERIC,
                                                                  "SERAP09T"                NUMERIC,
                                                                  "SERAP10P"                NUMERIC,
                                                                  "SERAP10T"                NUMERIC,
                                                                  "SERAP11P"                NUMERIC,
                                                                  "SERAP11T"                NUMERIC,
                                                                  "SERAPOIO"                NUMERIC,
                                                                  "RES_BIOL"                NUMERIC,
                                                                  "RES_QUIM"                NUMERIC,
                                                                  "RES_RADI"                NUMERIC,
                                                                  "RES_COMU"                NUMERIC,
                                                                  "COLETRES"                NUMERIC,
                                                                  "COMISS01"                NUMERIC,
                                                                  "COMISS02"                NUMERIC,
                                                                  "COMISS03"                NUMERIC,
                                                                  "COMISS04"                NUMERIC,
                                                                  "COMISS05"                NUMERIC,
                                                                  "COMISS06"                NUMERIC,
                                                                  "COMISS07"                NUMERIC,
                                                                  "COMISS08"                NUMERIC,
                                                                  "COMISS09"                NUMERIC,
                                                                  "COMISS10"                NUMERIC,
                                                                  "COMISS11"                NUMERIC,
                                                                  "COMISS12"                NUMERIC,
                                                                  "COMISSAO"                NUMERIC,
                                                                  "AP01CV01"                NUMERIC,
                                                                  "AP01CV02"                NUMERIC,
                                                                  "AP01CV05"                NUMERIC,
                                                                  "AP01CV06"                NUMERIC,
                                                                  "AP01CV03"                NUMERIC,
                                                                  "AP01CV04"                NUMERIC,
                                                                  "AP02CV01"                NUMERIC,
                                                                  "AP02CV02"                NUMERIC,
                                                                  "AP02CV05"                NUMERIC,
                                                                  "AP02CV06"                NUMERIC,
                                                                  "AP02CV03"                NUMERIC,
                                                                  "AP02CV04"                NUMERIC,
                                                                  "AP03CV01"                NUMERIC,
                                                                  "AP03CV02"                NUMERIC,
                                                                  "AP03CV05"                NUMERIC,
                                                                  "AP03CV06"                NUMERIC,
                                                                  "AP03CV03"                NUMERIC,
                                                                  "AP03CV04"                NUMERIC,
                                                                  "AP04CV01"                NUMERIC,
                                                                  "AP04CV02"                NUMERIC,
                                                                  "AP04CV05"                NUMERIC,
                                                                  "AP04CV06"                NUMERIC,
                                                                  "AP04CV03"                NUMERIC,
                                                                  "AP04CV04"                NUMERIC,
                                                                  "AP05CV01"                NUMERIC,
                                                                  "AP05CV02"                NUMERIC,
                                                                  "AP05CV05"                NUMERIC,
                                                                  "AP05CV06"                NUMERIC,
                                                                  "AP05CV03"                NUMERIC,
                                                                  "AP05CV04"                NUMERIC,
                                                                  "AP06CV01"                NUMERIC,
                                                                  "AP06CV02"                NUMERIC,
                                                                  "AP06CV05"                NUMERIC,
                                                                  "AP06CV06"                NUMERIC,
                                                                  "AP06CV03"                NUMERIC,
                                                                  "AP06CV04"                NUMERIC,
                                                                  "AP07CV01"                NUMERIC,
                                                                  "AP07CV02"                NUMERIC,
                                                                  "AP07CV05"                NUMERIC,
                                                                  "AP07CV06"                NUMERIC,
                                                                  "AP07CV03"                NUMERIC,
                                                                  "AP07CV04"                NUMERIC,
                                                                  "ATEND_PR"                NUMERIC,
                                                                  "NATJUR_ID"               VARCHAR(4));

                       /*Tabela dos tipos de retenção de tributos da mantenedora*/
                       CREATE TABLE IF NOT EXISTS {child_db}.codir("ID"                     VARCHAR(2),
                                                                   "RETENCAO"               VARCHAR(66));

                       /*Tabela dos tipos de órgão expedidor de alvará*/
                       CREATE TABLE IF NOT EXISTS {child_db}.orgexped("ID"                  VARCHAR(2),
                                                                      "EXPEDIDOR"           VARCHAR(66));

                       /*Tabela das classificações de avaliacao de estabelecimentos*/
                       CREATE TABLE IF NOT EXISTS {child_db}.clasaval("ID"                  VARCHAR(2),
                                                                      "AVALIACAO"           VARCHAR(66));

                       /*Tabelas únicas do sub bancos de dados cnes_dc*/

                       /*Tabela Dados Complementares (tabela principal do sub banco de dados cnes_dc)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.dcbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_DC"                   VARCHAR(2),
                                                                  "ANO_DC"                  INTEGER,
                                                                  "MES_DC"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "S_HBSAGP"                FLOAT,
                                                                  "S_HBSAGN"                FLOAT,
                                                                  "S_DPI"                   FLOAT,
                                                                  "S_DPAC"                  FLOAT,
                                                                  "S_REAGP"                 FLOAT,
                                                                  "S_REAGN"                 FLOAT,
                                                                  "S_REHCV"                 FLOAT,
                                                                  "MAQ_PROP"                FLOAT,
                                                                  "MAQ_OUTR"                FLOAT,
                                                                  "F_AREIA"                 NUMERIC,
                                                                  "F_CARVAO"                NUMERIC,
                                                                  "ABRANDAD"                NUMERIC,
                                                                  "DEIONIZA"                NUMERIC,
                                                                  "OSMOSE_R"                NUMERIC,
                                                                  "OUT_TRAT"                NUMERIC,
                                                                  "CNS_NEFR"                VARCHAR(15),
                                                                  "DIALISE"                 NUMERIC,
                                                                  "SIMUL_RD"                FLOAT,
                                                                   "PLANJ_RD"                FLOAT,
                                                                  "ARMAZ_FT"                FLOAT,
                                                                  "CONF_MAS"                FLOAT,
                                                                  "SALA_MOL"                FLOAT,
                                                                  "BLOCOPER"                FLOAT,
                                                                  "S_ARMAZE"                FLOAT,
                                                                  "S_PREPAR"                FLOAT,
                                                                  "S_QCDURA"                FLOAT,
                                                                  "S_QLDURA"                FLOAT,
                                                                  "S_CPFLUX"                FLOAT,
                                                                  "S_SIMULA"                FLOAT,
                                                                  "S_ACELL6"                FLOAT,
                                                                  "S_ALSEME"                FLOAT,
                                                                  "S_ALCOME"                FLOAT,
                                                                  "ORTV1050"                FLOAT,
                                                                  "ORV50150"                FLOAT,
                                                                  "OV150500"                FLOAT,
                                                                  "UN_COBAL"                FLOAT,
                                                                  "EQBRBAIX"                FLOAT,
                                                                  "EQBRMEDI"                FLOAT,
                                                                  "EQBRALTA"                FLOAT,
                                                                  "EQ_MAREA"                FLOAT,
                                                                  "EQ_MINDI"                FLOAT,
                                                                  "EQSISPLN"                FLOAT,
                                                                  "EQDOSCLI"                FLOAT,
                                                                  "EQFONSEL"                FLOAT,
                                                                  "CNS_ADM"                 VARCHAR(15),
                                                                  "CNS_OPED"                VARCHAR(15),
                                                                  "CNS_CONC"                VARCHAR(15),
                                                                  "CNS_OCLIN"               VARCHAR(15),
                                                                  "CNS_MRAD"                VARCHAR(15),
                                                                  "CNS_FNUC"                VARCHAR(15),
                                                                  "QUIMRADI"                NUMERIC,
                                                                  "S_RECEPC"                FLOAT,
                                                                  "S_TRIHMT"                FLOAT,
                                                                  "S_TRICLI"                FLOAT,
                                                                  "S_COLETA"                FLOAT,
                                                                  "S_AFERES"                FLOAT,
                                                                  "S_PREEST"                FLOAT,
                                                                  "S_PROCES"                FLOAT,
                                                                  "S_ESTOQU"                FLOAT,
                                                                  "S_DISTRI"                FLOAT,
                                                                  "S_SOROLO"                FLOAT,
                                                                  "S_IMUNOH"                FLOAT,
                                                                  "S_PRETRA"                FLOAT,
                                                                  "S_HEMOST"                FLOAT,
                                                                  "S_CONTRQ"                FLOAT,
                                                                  "S_BIOMOL"                FLOAT,
                                                                  "S_IMUNFE"                FLOAT,
                                                                  "S_TRANSF"                FLOAT,
                                                                  "S_SGDOAD"                FLOAT,
                                                                  "QT_CADRE"                FLOAT,
                                                                  "QT_CENRE"                FLOAT,
                                                                  "QT_REFSA"                FLOAT,
                                                                  "QT_CONRA"                FLOAT,
                                                                  "QT_EXTPL"                FLOAT,
                                                                  "QT_FRE18"                FLOAT,
                                                                  "QT_FRE30"                FLOAT,
                                                                  "QT_AGIPL"                FLOAT,
                                                                  "QT_SELAD"                FLOAT,
                                                                  "QT_IRRHE"                FLOAT,
                                                                  "QT_AGLTN"                FLOAT,
                                                                  "QT_MAQAF"                FLOAT,
                                                                  "QT_REFRE"                FLOAT,
                                                                  "QT_REFAS"                FLOAT,
                                                                  "QT_CAPFL"                FLOAT,
                                                                  "CNS_HMTR"                VARCHAR(15),
                                                                  "CNS_HMTL"                VARCHAR(15),
                                                                  "CNS_CRES"                VARCHAR(15),
                                                                  "CNS_RTEC"                VARCHAR(15),
                                                                  "HEMOTERA"                NUMERIC);

                       /*Tabelas únicas do sub bancos de dados cnes_pf*/

                       /*Tabela Profissionais (tabela principal do sub banco de dados cnes_pf)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.pfbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_PF"                   VARCHAR(2),
                                                                  "ANO_PF"                  INTEGER,
                                                                  "MES_PF"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "CBO_ID"                  VARCHAR(6),
                                                                  "CBOUNICO_ID"             VARCHAR(6),
                                                                  "NOMEPROF"                VARCHAR(60),
                                                                  "CNS_PROF"                VARCHAR(15),
                                                                  "CONSELHO_ID"             VARCHAR(2),
                                                                  "REGISTRO"                VARCHAR(13),
                                                                  "VINCULAC_ID"             VARCHAR(6),
                                                                  "VINCUL_C"                NUMERIC,
                                                                  "VINCUL_A"                NUMERIC,
                                                                  "VINCUL_N"                NUMERIC,
                                                                  "PROF_SUS"                NUMERIC,
                                                                  "PROFNSUS"                NUMERIC,
                                                                  "HORAOUTR"                FLOAT,
                                                                  "HORAHOSP"                FLOAT,
                                                                  "HORA_AMB"                FLOAT);

                       /*Tabela das especialidades dos profissionais*/
                       CREATE TABLE IF NOT EXISTS {child_db}.cbo("ID"                       VARCHAR(6),
                                                                 "OCUPACAO"                 VARCHAR(66));

                       /*Em verdade é a mesma coisa da tabela anterior*/
                       CREATE TABLE IF NOT EXISTS {child_db}.cbounico("ID"                  VARCHAR(6),
                                                                      "OCUPACAO"            VARCHAR(66));

                       /*Tabeça dos conselhos de profissão*/
                       CREATE TABLE IF NOT EXISTS {child_db}.conselho("ID"                  VARCHAR(2),
                                                                      "DENOMINACAO"         VARCHAR(66));

                       /*Tabeça dos tipos de vínculos do profissional*/
                       CREATE TABLE IF NOT EXISTS {child_db}.vinculac("ID"                  VARCHAR(6),
                                                                      "DESCRICAO"           VARCHAR(100));

                       /*Tabelas únicas do sub bancos de dados cnes_lt*/

                       /*Tabela Leitos (tabela principal do sub banco de dados cnes_lt)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.ltbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_LT"                   VARCHAR(2),
                                                                  "ANO_LT"                  INTEGER,
                                                                  "MES_LT"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "TPLEITO_ID"              VARCHAR(2),
                                                                  "CODLEITO_ID"             VARCHAR(2),
                                                                  "QT_EXIST"                FLOAT,
                                                                  "QT_CONTR"                FLOAT,
                                                                  "QT_SUS"                  FLOAT);

                       /*Tabela dos tipos de leito do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.tpleito("ID"                   VARCHAR(2),
                                                                     "TIPO"                 VARCHAR(66));

                       /*Tabela dos tipos específicos dos leitos*/
                       CREATE TABLE IF NOT EXISTS {child_db}.codleito("ID"                  VARCHAR(2),
                                                                      "ESPECIALIDADE"       VARCHAR(66));

                       /*Tabelas únicas do sub bancos de dados cnes_eq*/

                       /*Tabela Equipamentos (tabela principal do sub banco de dados cnes_eq)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.eqbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_EQ"                   VARCHAR(2),
                                                                  "ANO_EQ"                  INTEGER,
                                                                  "MES_EQ"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "TIPEQUIP_ID"             VARCHAR(2),
                                                                  "CODEQUIP_ID"             VARCHAR(2),
                                                                  "QT_EXIST"                FLOAT,
                                                                  "QT_USO"                  FLOAT,
                                                                  "IND_SUS"                 NUMERIC,
                                                                  "ND_NSUS"                 NUMERIC);

                       /*Tabela dos tipos de equipamentos*/
                       CREATE TABLE IF NOT EXISTS {child_db}.tipequip("ID"                  VARCHAR(2),
                                                                      "TIPO"                VARCHAR(66));

                       /*Tabela dos nomes dos equipamentos*/
                       CREATE TABLE IF NOT EXISTS {child_db}.codequip("ID"                  VARCHAR(2),
                                                                      "DENOMINACAO"         VARCHAR(66));


                       /*Tabelas únicas do sub bancos de dados cnes_sr*/

                       /*Tabela Serviços Especializados (tabela principal do sub banco de dados cnes_sr)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.srbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_SR"                   VARCHAR(2),
                                                                  "ANO_SR"                  INTEGER,
                                                                  "MES_SR"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "SERVESP_ID"              VARCHAR(3),
                                                                  "CLASSSR_ID"              VARCHAR(6),
                                                                  "SRVUNICO_ID"             VARCHAR(3),
                                                                  "TPGESTAO_ID"             VARCHAR(2),
                                                                  "PFPJ_ID"                 VARCHAR(2),
                                                                  "CPF_CNPJ"                VARCHAR(14),
                                                                  "NIVDEP_ID"               VARCHAR(2),
                                                                  "ESFERAA_ID"              VARCHAR(2),
                                                                  "ATIVIDAD_ID"             VARCHAR(2),
                                                                  "RETENCAO_ID"             VARCHAR(2),
                                                                  "NATUREZA_ID"             VARCHAR(2),
                                                                  "CLIENTEL_ID"             VARCHAR(2),
                                                                  "TPUNID_ID"               VARCHAR(2),
                                                                  "TURNOAT_ID"              VARCHAR(2),
                                                                  "NIVHIER_ID"              VARCHAR(2),
                                                                  "TERCEIRO"                NUMERIC,
                                                                  "CNPJ_MAN"                VARCHAR(14),
                                                                  "CARACTER_ID"             VARCHAR(2),
                                                                  "AMB_NSUS"                NUMERIC,
                                                                  "AMB_SUS"                 NUMERIC,
                                                                  "HOSP_NSUS"               NUMERIC,
                                                                  "HOSP_SUS"                NUMERIC,
                                                                  "CONTSRVU"                NUMERIC,
                                                                  "CNESTERC"                VARCHAR(7),
                                                                  "NATJUR_ID"               VARCHAR(4));

                       /*Tabela dos serviços especializados*/
                       CREATE TABLE IF NOT EXISTS {child_db}.servesp("ID"                   VARCHAR(3),
                                                                     "DESCRICAO"            VARCHAR(100));

                       /*Tabela de classificações dos serviços especializados*/
                       CREATE TABLE IF NOT EXISTS {child_db}.classsr("ID"                   VARCHAR(6),
                                                                     "DESCRICAO"            VARCHAR(100));

                       /*Na verdade é a mesma coisa da tabela servesp*/
                       CREATE TABLE IF NOT EXISTS {child_db}.srvunico("ID"                  VARCHAR(3),
                                                                      "DESCRICAO"           VARCHAR(100));

                       /*Tabela de caracterizações do estabelecimento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.caracter("ID"                  VARCHAR(2),
                                                                      "CARACTERIZACAO"      VARCHAR(66));

                       /*Tabelas únicas do sub bancos de dados cnes_ep*/

                       /*Tabela Equipes (tabela principal do sub banco de dados cnes_ep)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.epbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_EP"                   VARCHAR(2),
                                                                  "ANO_EP"                  INTEGER,
                                                                  "MES_EP"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "IDEQUIPE_ID"             VARCHAR(18),
                                                                  "TIPOEQP_ID"              VARCHAR(2),
                                                                  "NOME_EQP"                VARCHAR(60),
                                                                  "IDAREA_ID"               VARCHAR(10),
                                                                  "NOMEAREA"                VARCHAR(60),
                                                                  "IDSEGM_ID"               VARCHAR(8),
                                                                  "DESCSEGM"                VARCHAR(60),
                                                                  "TIPOSEGM_ID"             VARCHAR(2),
                                                                  "DT_ATIVA"                DATE,
                                                                  "DT_DESAT"                DATE,
                                                                  "QUILOMBO"                NUMERIC,
                                                                  "ASSENTAD"                NUMERIC,
                                                                  "POPGERAL"                NUMERIC,
                                                                  "ESCOLA"                  NUMERIC,
                                                                  "INDIGENA"                NUMERIC,
                                                                  "PRONASCI"                NUMERIC,
                                                                  "MOTDESAT_ID"             VARCHAR(2),
                                                                  "TPDESAT_ID"              VARCHAR(2));

                       /*Tabelas da identificação de equipes*/
                       CREATE TABLE IF NOT EXISTS {child_db}.idequipe("ID"                  VARCHAR(18),
                                                                      "NOME_EQUIPE"         VARCHAR(66));

                       /*Tabela dos tipos de equipe*/
                       CREATE TABLE IF NOT EXISTS {child_db}.tipoeqp("ID"                   VARCHAR(2),
                                                                     "TIPO"                 VARCHAR(66));

                       /*Tabela da identificação de áreas*/
                       CREATE TABLE IF NOT EXISTS {child_db}.idarea("ID"                    VARCHAR(10),
                                                                    "NOME_AREA"             VARCHAR(66));

                       /*Tabela da identificação de segmentos*/
                       CREATE TABLE IF NOT EXISTS {child_db}.idsegm("ID"                    VARCHAR(8),
                                                                    "DESCRICAO"             VARCHAR(66));

                       /*Tabela dos tipos de segmento*/
                       CREATE TABLE IF NOT EXISTS {child_db}.tiposegm("ID"                  VARCHAR(2),
                                                                      "TIPO"                VARCHAR(66));

                       /*Tabela dos motivos de desativação*/
                       CREATE TABLE IF NOT EXISTS {child_db}.motdesat("ID"                  VARCHAR(2),
                                                                      "MOTIVO"              VARCHAR(66));

                       /*Tabela dos tipos de desativação*/
                       CREATE TABLE IF NOT EXISTS {child_db}.tpdesat("ID"                   VARCHAR(2),
                                                                     "TIPO"                 VARCHAR(66));

                       /*Tabelas únicas do sub bancos de dados cnes_hb*/

                       /*Tabela Habilitações (tabela principal do sub banco de dados cnes_hb)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.hbbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_HB"                   VARCHAR(2),
                                                                  "ANO_HB"                  INTEGER,
                                                                  "MES_HB"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "TERCEIRO"                NUMERIC,
                                                                  "SGRUPHAB_ID"             VARCHAR(4),
                                                                  "CMPT_INI"                DATE,
                                                                  "CMPT_FIM"                DATE,
                                                                  "DTPORTAR"                DATE,
                                                                  "PORTARIA"                VARCHAR(50),
                                                                  "MAPORTAR"                DATE,
                                                                  "NULEITOS"                FLOAT);

                       /*Tabela dos tipos de habilitação*/
                       CREATE TABLE IF NOT EXISTS {child_db}.sgruphab_hb("ID"               VARCHAR(4),
                                                                         "HABILITACAO"      VARCHAR(100));

                       /*Tabelas únicas do sub bancos de dados cnes_rc*/

                       /*Tabela Regras Contratuais (tabela principal do sub banco de dados cnes_rc)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.rcbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_RC"                   VARCHAR(2),
                                                                  "ANO_RC"                  INTEGER,
                                                                  "MES_RC"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "TERCEIRO"                NUMERIC,
                                                                  "SGRUPHAB_ID"             VARCHAR(4),
                                                                  "CMPT_INI"                DATE,
                                                                  "CMPT_FIM"                DATE,
                                                                  "DTPORTAR"                DATE,
                                                                  "PORTARIA"                VARCHAR(50),
                                                                  "MAPORTAR"                DATE);

                       /*Tabela dos tipos de regra contratual*/
                       CREATE TABLE IF NOT EXISTS {child_db}.sgruphab_rc("ID"               VARCHAR(4),
                                                                         "REGRA"            VARCHAR(150));

                       /*Tabelas únicas do sub bancos de dados cnes_gm*/

                       /*Tabela Gestão e Metas (tabela principal do sub banco de dados cnes_gm)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.gmbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_GM"                   VARCHAR(2),
                                                                  "ANO_GM"                  INTEGER,
                                                                  "MES_GM"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "TERCEIRO"                NUMERIC,
                                                                  "SGRUPHAB_ID"             VARCHAR(4),
                                                                  "CMPT_INI"                DATE,
                                                                  "CMPT_FIM"                DATE,
                                                                  "DTPORTAR"                DATE,
                                                                  "PORTARIA"                VARCHAR(50),
                                                                  "MAPORTAR"                DATE);

                       /*Tabela dos tipos de gestão*/
                       CREATE TABLE IF NOT EXISTS {child_db}.sgruphab_gm("ID"               VARCHAR(4),
                                                                         "GESTAO"           VARCHAR(100));

                       /*Tabelas únicas do sub bancos de dados cnes_ee*/

                       /*Tabela Estabelecimentos de Ensino (tabela principal do sub banco de dados cnes_ee)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.eebr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_EE"                   VARCHAR(2),
                                                                  "ANO_EE"                  INTEGER,
                                                                  "MES_EE"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "TPGESTAO_ID"             VARCHAR(2),
                                                                  "PFPJ_ID"                 VARCHAR(2),
                                                                  "CPF_CNPJ"                VARCHAR(14),
                                                                  "NIVDEP_ID"               VARCHAR(2),
                                                                  "CNPJ_MAN"                VARCHAR(14),
                                                                  "ESFERAA_ID"              VARCHAR(2),
                                                                  "RETENCAO_ID"             VARCHAR(2),
                                                                  "ATIVIDAD_ID"             VARCHAR(2),
                                                                  "NATUREZA_ID"             VARCHAR(2),
                                                                  "CLIENTEL_ID"             VARCHAR(2),
                                                                  "TPUNID_ID"               VARCHAR(2),
                                                                  "TURNOAT_ID"              VARCHAR(2),
                                                                  "NIVHIER_ID"              VARCHAR(2),
                                                                  "TERCEIRO"                NUMERIC,
                                                                  "COD_CEP"                 VARCHAR(8),
                                                                  "VINC_SUS"                NUMERIC,
                                                                  "TPPREST_ID"              VARCHAR(2),
                                                                  "SGRUPHAB_ID"             VARCHAR(4),
                                                                  "CMPT_INI"                DATE,
                                                                  "CMPT_FIM"                DATE,
                                                                  "DTPORTAR"                DATE,
                                                                  "PORTARIA"                VARCHAR(50),
                                                                  "MAPORTAR"                DATE,
                                                                  "NATJUR_ID"               VARCHAR(4));

                       /*Tabela dos tipos de estabelecimento (só o próprio de ensino na verdade)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.sgruphab_ee("ID"               VARCHAR(4),
                                                                         "TIPO"             VARCHAR(66));

                       /*Tabelas únicas do sub bancos de dados cnes_ef*/

                       /*Tabela Estabelecimentos Filantrópicos (tabela principal do sub banco de dados cnes_ef)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.efbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_EF"                   VARCHAR(2),
                                                                  "ANO_EF"                  INTEGER,
                                                                  "MES_EF"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "TPGESTAO_ID"             VARCHAR(2),
                                                                  "PFPJ_ID"                 VARCHAR(2),
                                                                  "CPF_CNPJ"                VARCHAR(14),
                                                                  "NIVDEP_ID"               VARCHAR(2),
                                                                  "CNPJ_MAN"                VARCHAR(14),
                                                                  "ESFERAA_ID"              VARCHAR(2),
                                                                  "RETENCAO_ID"             VARCHAR(2),
                                                                  "ATIVIDAD_ID"             VARCHAR(2),
                                                                  "NATUREZA_ID"             VARCHAR(2),
                                                                  "CLIENTEL_ID"             VARCHAR(2),
                                                                  "TPUNID_ID"               VARCHAR(2),
                                                                  "TURNOAT_ID"              VARCHAR(2),
                                                                  "NIVHIER_ID"              VARCHAR(2),
                                                                  "TERCEIRO"                NUMERIC,
                                                                  "COD_CEP"                 VARCHAR(8),
                                                                  "VINC_SUS"                NUMERIC,
                                                                  "TPPREST_ID"              VARCHAR(2),
                                                                  "SGRUPHAB_ID"             VARCHAR(4),
                                                                  "CMPT_INI"                DATE,
                                                                  "CMPT_FIM"                DATE,
                                                                  "DTPORTAR"                DATE,
                                                                  "PORTARIA"                VARCHAR(50),
                                                                  "MAPORTAR"                DATE,
                                                                  "NATJUR_ID"               VARCHAR(4));

                       /*Tabela dos tipos de estabelecimento (só o próprio filantrópico na verdade)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.sgruphab_ef("ID"               VARCHAR(4),
                                                                         "TIPO"             VARCHAR(66));

                       /*Tabelas únicas do sub bancos de dados cnes_in*/

                       /*Tabela Incentivos (tabela principal do sub banco de dados cnes_in)*/
                       CREATE TABLE IF NOT EXISTS {child_db}.inbr("CNES_ID"                 VARCHAR(7),
                                                                  "UF_IN"                   VARCHAR(2),
                                                                  "ANO_IN"                  INTEGER,
                                                                  "MES_IN"                  VARCHAR(2),
                                                                  "CODUFMUN_ID"             VARCHAR(6),
                                                                  "TPGESTAO_ID"             VARCHAR(2),
                                                                  "PFPJ_ID"                 VARCHAR(2),
                                                                  "CPF_CNPJ"                VARCHAR(14),
                                                                  "NIVDEP_ID"               VARCHAR(2),
                                                                  "CNPJ_MAN"                VARCHAR(14),
                                                                  "ESFERAA_ID"              VARCHAR(2),
                                                                  "RETENCAO_ID"             VARCHAR(2),
                                                                  "ATIVIDAD_ID"             VARCHAR(2),
                                                                  "NATUREZA_ID"             VARCHAR(2),
                                                                  "CLIENTEL_ID"             VARCHAR(2),
                                                                  "TPUNID_ID"               VARCHAR(2),
                                                                  "TURNOAT_ID"              VARCHAR(2),
                                                                  "NIVHIER_ID"              VARCHAR(2),
                                                                  "TERCEIRO"                NUMERIC,
                                                                  "COD_CEP"                 VARCHAR(8),
                                                                  "VINC_SUS"                NUMERIC,
                                                                  "TPPREST_ID"              VARCHAR(2),
                                                                  "SGRUPHAB"                VARCHAR(4),
                                                                  "CMPT_INI"                DATE,
                                                                  "CMPT_FIM"                DATE,
                                                                  "DTPORTAR"                DATE,
                                                                  "PORTARIA"                VARCHAR(50),
                                                                  "MAPORTAR"                DATE,
                                                                  "NATJUR_ID"               VARCHAR(4));
                ''')
    conn.commit()
    # Encerra o cursor
    cursor.close()
    # Encerra a conexão
    conn.close()



"""
Módulo que contém funções que leem arquivo "dbc" em plataformas Windows utilizando o programa "dbf2dbc.exe"
e arquivo texto em formato "cnv".
"""

import os
import subprocess
from pathlib import Path

from dbfread import DBF
import pandas as pd


def dbc2dbf(infile):
    """
    Descompacta um arquivo "dbc" para "dbf" utilizando o programa executável "dbf2dbc" pertencente ao Datasus.

    Parâmetros
    ----------
    infile: objeto str
        String do nome do arquivo "dbc"

    Retorno
    -------
    outfile: objeto str
        String do nome do arquivo "dbf"
    """

    subprocess.run(['C:/dbf2dbc/dbf2dbc.exe', infile])
    os.unlink(infile)
    outfile = infile[:-4] + '.dbf'

    return outfile


def read_dbc(filename, signature='utf-8'):
    """
    Descompacta um arquivo "dbc" para "dbf", em seguida o lê como tal e por fim o converte em um objeto
    pandas DataFrame e o elimina.

    Parâmetros
    ----------
    filename: objeto str
        String do nome do arquivo "dbc"
    signature: objeto str
        String do nome do formato de encoding do arquivo "dbc"

    Retorno
    -------
    df: objeto pandas DataFrame
        Dataframe que contém os dados de um arquivo principal de dados originalmente em formato "dbc"
    """

    file_name = dbc2dbf(filename)
    dbf = DBF(file_name, encoding=signature)
    df = pd.DataFrame(iter(dbf))
    os.unlink(file_name)

    return df


def attempt_int(value):
    """
    Converte um objeto referenciado pelo parâmetro "value" para o tipo int ou caso não seja possível
    o retorna íntegro.

    Parâmetros
    ----------
    infile: objeto de qualquer tipo
        Objeto que se deseja converter para o tipo int

    Retorno
    -------
    df: objeto int ou na sua class original
        Objeto convertido para o tipo int ou íntegro
    """

    try:
        return int(value)
    except (ValueError, TypeError):
        return value


def read_cnv(filename):
    """
    Converte o conteúdo útil do arquivo texto em formato "cnv" para um objeto pandas DataFrame.

    Parâmetros
    ----------
    filename: objeto str
        String do nome do arquivo "cnv"

    Retorno
    -------
    df: objeto pandas DataFrame
        Dataframe que contém os dados úteis de um arquivo auxiliar de dados em formato "cnv"
    """

    # Cria um file handler do arquivo "cnv"
    fhandler = open(filename)

    k = 0  # Inicialização de uma variável para identificar a primeira e as outras linhas
    lista_significacao = []  # "Inicialização" da lista de decodificações
    lista_id = []  # "Inicialização" da lista de códigos
    # Leitura do arquivo texto
    if filename[:-4] in ['SINAN_Classdeng', 'cnv/CNES_NATUREZA']:   # Estes são arquivos "cnv" cujos números de algarismos dos códigos estão na segunda linha do arquivo
        for linha in fhandler:
            k += 1
            linha = linha.rstrip()
            # Passa a primeira linha
            if k == 1:
                continue
                # Coleta na segunda linha do arquivo o número de caracteres dos códigos
            elif k == 2:
                size_id = int(linha.split()[1])
            # Coleta nas outras linhas do arquivo a decodificação e o respectivo código, respectivamente, e append o elemento coletado às respectivas listas
            else:
                lista_significacao.append(linha[9:60].strip())
                lista_id.append(linha[60:60+size_id].strip())

    elif filename[:-4] in ['cnv/CNES_TP_ESTAB', 'cnv/CNES_NATJUR', 'SIH_LEITOS']:  # Estes são arquivos "cnv" cujas significações não começam na posição (python) 9...
                                                                                   # (mas na 11), cujos códigos não começam na posição (python) 60 (mas na 112) de...
                                                                                   # cada linha do arquivo, e cujos números de algarismos dos códigos não são o segundo...
                                                                                   # elemento listado na primeira linha (mas o terceiro)
        for linha in fhandler:
            k += 1
            linha = linha.rstrip()
            # Coleta na primeira linha do arquivo o número de caracteres dos códigos
            if k == 1:
                size_id = int(linha.split()[2])
            # Coleta nas outras linhas do arquivo a decodificação e o respectivo código, respectivamente, e append o elemento coletado às respectivas listas
            else:
                lista_significacao.append(linha[11:112].strip())
                lista_id.append(linha[112:112+size_id].strip())

    elif filename[:-4] == 'cnv/CNES_Equip_Tp':  # Os dois últimos dígitos da coluna de códigos desse arquivo permite construir a parent table CODEQUIP e assim...
                                            # decodificar a coluna CODEQUIP_ID da tabela eqbr do banco de dados CNES_EQ
        size_id = 2
        for linha in fhandler:
            k += 1
            linha = linha.rstrip()
            # O número de caracteres dos códigos já foi definido acima
            if k == 1:
                continue
            # Coleta nas outras linhas do arquivo a decodificação e o respectivo código, respectivamente, e append o elemento coletado às respectivas listas
            else:
                lista_significacao.append(linha[9:60].strip())
                lista_id.append(linha[61:61+size_id].strip())

    elif filename[:-4] == 'SIA_CARAT_AT': # Estes são arquivos "cnv" cujos números de algarismos dos códigos não estão na terceira linha do arquivo
        for linha in fhandler:
            k += 1
            linha = linha.rstrip()
            # Passa a primeira linha
            if (k == 1) or (k == 2):
                continue
                # Coleta na terceira linha do arquivo o número de caracteres dos códigos
            elif k == 3:
                size_id = int(linha.split()[1])
            # Coleta nas outras linhas do arquivo a decodificação e o respectivo código, respectivamente, e append o elemento coletado às respectivas listas
            else:
                lista_significacao.append(linha[9:60].strip())
                lista_id.append(linha[60:60+size_id].strip())

    else: # Arquivos "cnv" padrão
        for linha in fhandler:
            k += 1
            linha = linha.rstrip()
            # Coleta na primeira linha do arquivo o número de caracteres dos códigos
            if k == 1:
                size_id = int(linha.split()[1])
            # Coleta nas outras linhas do arquivo a decodificação e o respectivo código, respectivamente, e append o elemento coletado às respectivas listas
            else:
                lista_significacao.append(linha[9:60].strip())
                lista_id.append(linha[60:60+size_id].strip())

    # "Inicialização" de um objeto pandas DataFrame com as colunas ID e SIGNIFICACAO
    df = pd.DataFrame(columns=['ID', 'SIGNIFICACAO'])
    # Transferência dos dados da lista de códigos para a coluna ID do objeto pandas DataFrame
    df['ID'] = lista_id
    # Transferência dos dados da lista de decodificações para a coluna SIGNIFICACAO do objeto pandas DataFrame
    df['SIGNIFICACAO'] = lista_significacao
    # Remove o arquivo "cnv" após fechar o "fhandler"
    fhandler.close()
    os.remove(filename)
    # Torna UPPER CASE todos os valores da coluna SIGNIFICACAO
    df['SIGNIFICACAO'] = df['SIGNIFICACAO'].apply(lambda x: x.upper())
    # Modifica o valor tipo string da célula da coluna SIGNIFICACAO que contém o valor Ign
    df['SIGNIFICACAO'].replace('IGN', 'IGNORADO', inplace=True)
    df = df.drop(df[df['SIGNIFICACAO']=='IGNORADO'].index)
    # Elimina eventual linha com string impertinente na coluna ID
    df = df.drop(df[df['ID']==''].index)
    df = df.drop(df[df['ID']==','].index)

    # Elimina duplicates do objeto pandas DataFrame com base na coluna ID
    df.drop_duplicates(subset='ID', keep='first', inplace=True)

    if (len(df.loc[df['ID'].astype(str).str.isdigit(), 'ID'].tolist()) == df.shape[0]) or (len(df.loc[~df['ID'].astype(str).str.isdigit(), 'ID'].tolist()) == df.shape[0]):
        # Converte de string para int cada valor da coluna de códigos do objeto pandas DataFrame se possível
        df['ID'] = df['ID'].apply(lambda x: attempt_int(x))
        # Ordena o objeto pandas DataFrame com base na coluna ID
        df.sort_values(by=['ID'], inplace=True)

    # Converte de int para string cada valor da coluna de códigos do objeto pandas DataFrame
    df['ID'] = df['ID'].apply(lambda x: str(x))

    # Reset o index do objeto pandas DataFrame devido ao sorting e ao droping evetualmente efetuados
    df.reset_index(drop=True, inplace=True)
    return df



###########################################################################################################################
#  SINASC SINASC SINASC SINASC SINASC SINASC SINASC SINASC SINASC SINASC SINASC SINASC SINASC SINASC SINASC SINASC SINASC #
###########################################################################################################################

import os
import time
from datetime import datetime

import numpy as np
import pandas as pd
import psycopg2

from transform.prepare_SINASC import DataSinascMain, DataSinascAuxiliary

###########################################################################################################################
#  pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas #
###########################################################################################################################
###########################################################################################################################
#     AUXILIARY TABLES * AUXILIARY TABLES * AUXILIARY TABLES * AUXILIARY TABLES * AUXILIARY TABLES * AUXILIARY TABLES     #
###########################################################################################################################
# Função que utiliza "pandas.to_sql" para a inserção de dados não principais no banco de dados "child_db"
def insert_into_most_SINASC_tables(path, device, child_db):

    label1 = 'append'
    label2 = 'ID'

    # Cria uma instância da classe "DataSinascAuxiliary" do módulo "prepare_SINASC" do package "data_wrangling"
    data_sinasc_auxiliary = DataSinascAuxiliary(path)

    # Chama métodos da classe "DataSinascAuxiliary" do módulo "prepare_SINASC" do banco de dados SINASC
    df_CNESDN = data_sinasc_auxiliary.get_CNESDN_treated()
    df_CNESDN.to_sql('codestab', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TABUF = data_sinasc_auxiliary.get_TABUF_treated()
    df_TABUF.to_sql('ufcod', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_RSAUDE = data_sinasc_auxiliary.get_RSAUDE_treated()
    df_RSAUDE.to_sql('rsaude', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_CADMUN = data_sinasc_auxiliary.get_CADMUN_treated()
    df_CADMUN.to_sql('codmunnasc', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_LOCOCOR = data_sinasc_auxiliary.get_LOCOCOR_treated()
    df_LOCOCOR.to_sql('locnasc', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_SITCONJU = data_sinasc_auxiliary.get_SITCONJU_treated()
    df_SITCONJU.to_sql('estcivmae', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_INSTRUC = data_sinasc_auxiliary.get_INSTRUC_treated()
    df_INSTRUC.to_sql('escmae', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TABOCUP_2TCC = data_sinasc_auxiliary.get_TABOCUP_2TCC_treated()
    df_TABOCUP_2TCC.to_sql('codocupmae', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_SEMANAS = data_sinasc_auxiliary.get_SEMANAS_treated()
    df_SEMANAS.to_sql('gestacao', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    # Mesmo objeto pandas DataFrame da tabela "codmunnasc"
    df_CADMUN.to_sql('codmunres', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_GRAVIDEZ = data_sinasc_auxiliary.get_GRAVIDEZ_treated()
    df_GRAVIDEZ.to_sql('gravidez', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_PARTO = data_sinasc_auxiliary.get_PARTO_treated()
    df_PARTO.to_sql('parto', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_CONSULT = data_sinasc_auxiliary.get_CONSULT_treated()
    df_CONSULT.to_sql('consultas', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_RACA = data_sinasc_auxiliary.get_RACA_treated()
    df_RACA.to_sql('racacor', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_CID10 = data_sinasc_auxiliary.get_CID10_treated()
    df_CID10.to_sql('codanomal', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_NAT1212 = data_sinasc_auxiliary.get_NAT1212_treated()
    df_NAT1212.to_sql('naturalmae', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    # Mesmo objeto pandas DataFrame da tabela "codmunnasc"
    df_CADMUN.to_sql('codmunnatu', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_ESC2010 = data_sinasc_auxiliary.get_ESC2010_treated()
    df_ESC2010.to_sql('escmae2010', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    # Mesmo objeto pandas DataFrame da tabela "racacor"
    df_RACA.to_sql('racacormae', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TPMETODO = data_sinasc_auxiliary.get_TPMETODO_treated()
    df_TPMETODO.to_sql('tpmetestim', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TPAPRESENT = data_sinasc_auxiliary.get_TPAPRESENT_treated()
    df_TPAPRESENT.to_sql('tpapresent', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_STTRABPART = data_sinasc_auxiliary.get_STTRABPART_treated()
    df_STTRABPART.to_sql('sttrabpart', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_STPARTO = data_sinasc_auxiliary.get_STPARTO_treated()
    df_STPARTO.to_sql('stcesparto', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TPASSIST = data_sinasc_auxiliary.get_TPASSIST_treated()
    df_TPASSIST.to_sql('tpnascassi', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TPFUNCRESP = data_sinasc_auxiliary.get_TPFUNCRESP_treated()
    df_TPFUNCRESP.to_sql('tpfuncresp', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_ESCAGR1 = data_sinasc_auxiliary.get_ESCAGR1_treated()
    df_ESCAGR1.to_sql('escmaeagr1', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TABPAIS = data_sinasc_auxiliary.get_TABPAIS_treated()
    df_TABPAIS.to_sql('codpaisres', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_ROBSON = data_sinasc_auxiliary.get_ROBSON_treated()
    df_ROBSON.to_sql('tprobson', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)


###########################################################################################################################
#    copy_expert+pandas copy_expert+pandas copy_expert+pandas copy_expert+pandas copy_expert+pandas copy_expert+pandas    #
###########################################################################################################################
###########################################################################################################################
#    MAIN TABLE * MAIN TABLE * MAIN TABLE * MAIN TABLE * MAIN TABLE * MAIN TABLE * MAIN TABLE * MAIN TABLE * MAIN TABLE   #
###########################################################################################################################
# Função que utiliza "copy_expert" para a inserção de dados principais e "pandas.to_sql" para a inserção
# dos respectivos metadados no banco de dados "child_db"
def insert_into_main_table_and_arquivos(file_name, directory, date_ftp, device, child_db, connection_data):
    start = time.time()
    counting_rows = pd.read_sql(f'''SELECT COUNT('NOME') FROM {child_db}.arquivos''', con=device)
    qtd_files_pg = counting_rows.iloc[0]['count']
    print(f'A quantidade de arquivos principais de dados do {child_db} já carregada no {connection_data[0]}/PostgreSQL é {qtd_files_pg}.')

    # Tratamento de dados principais do SINASC
    base = file_name[0:2]
    state = file_name[2:4]
    year = file_name[4:8]
    main_table = base.lower() + 'br'

    print(f'\nIniciando a lida com o arquivo DN{state}{year}...')
    # Cria uma instância da classe "DataSinascMain" do módulo "prepare_SINASC" do package "data_wrangling"
    data_sinasc_main = DataSinascMain(state, year)
    # Chama método da classe "DataSinascMain" do módulo "prepare_SINASC" referentes ao banco de dados sinasc
    df = data_sinasc_main.get_DNXXaaaa_treated()

    # Inserção das colunas UF_DN e ANO_DN no objeto pandas DataFrame "df"
    df.insert(1, 'UF_' + base, [state]*df.shape[0])
    df.insert(2, 'ANO_' + base, [int(year)]*df.shape[0])

    # Criação de arquivo "csv" contendo os dados do arquivo principal de dados do sinasc armazenado no objeto...
    # pandas DataFrame "df"
    df.to_csv(base + state + year + '.csv', sep=',', header=False, index=False, escapechar=' ')
    # Leitura do arquivo "csv" contendo os dados do arquivo principal de dados do sinasc
    f = open(base + state + year + '.csv', 'r')
    # Conecta ao banco de dados mãe "connection_data[0]" do SGBD PostgreSQL usando o módulo python "psycopg2"
    conn = psycopg2.connect(dbname=connection_data[0],
                            host=connection_data[1],
                            port=connection_data[2],
                            user=connection_data[3],
                            password=connection_data[4])
    # Criação de um cursor da conexão tipo "psycopg2" referenciado à variável "cursor"
    cursor = conn.cursor()
    try:
        # Faz a inserção dos dados armazenados em "f" na tabela "main_table" do banco de dados "child_db"...
        # usando o método "copy_expert" do "psycopg2"
        cursor.copy_expert(f'''COPY {child_db}.{main_table} FROM STDIN WITH CSV DELIMITER AS ',';''', f)
    except:
        print(f'Tentando a inserção do arquivo {base}{state}{year} por método alternativo (pandas)...')
        df.to_sql(main_table, con=device, schema=child_db, if_exists='append', index=False)
    else:
        conn.commit()
    # Encerra o cursor
    cursor.close()
    # Encerra a conexão
    conn.close()
    # Encerra o file handler
    f.close()
    # Remoção do arquivo "csv"
    os.remove(base + state + year + '.csv')
    print(f'Terminou de inserir os dados do arquivo {base}{state}{year} na tabela {main_table} do banco de dados {child_db}.')

    # Cria um objeto pandas DataFrame com apenas uma linha de dados, a qual contém informações sobre o...
    # arquivo de dados principal carregado
    file_data = pd.DataFrame(data=[[file_name, directory, date_ftp, datetime.today(), int(df.shape[0])]],
                             columns= ['NOME', 'DIRETORIO', 'DATA_INSERCAO_FTP', 'DATA_HORA_CARGA', 'QTD_REGISTROS'],
                             index=None
                             )
    # Inserção de informações do arquivo principal de dados no banco de dados "child_db"
    file_data.to_sql('arquivos', con=device, schema=child_db, if_exists='append', index=False)
    print(f'Terminou de inserir os metadados do arquivo {base}{state}{year} na tabela arquivos do banco de dados {child_db}.')
    end = time.time()
    print(f'Demorou {round((end - start)/60, 1)} minutos para essas duas inserções no {connection_data[0]}/PostgreSQL!')



###########################################################################################################################
# SIA_ANY SIA_ANY SIA_ANY SIA_ANY SIA_ANY SIA_ANY SIA_ANY SIA_ANY SIA_ANY SIA_ANY SIA_ANY SIA_ANY SIA_ANY SIA_ANY SIA_ANY #
###########################################################################################################################

import os
import time
from datetime import datetime

import numpy as np
import pandas as pd
import psycopg2

from transform.prepare_SIA import DataSiaMain, DataSiaAuxiliary

###########################################################################################################################
#  pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas pandas #
###########################################################################################################################
###########################################################################################################################
#     AUXILIARY TABLES * AUXILIARY TABLES * AUXILIARY TABLES * AUXILIARY TABLES * AUXILIARY TABLES * AUXILIARY TABLES     #
###########################################################################################################################
# Função que utiliza para a inserção de dados não principais o pandas.to_sql + SQLAlchemy
def insert_into_most_SIA_PA_tables(path, device, child_db):

    label1 = 'append'
    label2 = 'ID'

    # Cria uma instância da classe "DataSiaAuxiliary" do módulo "prepare_SIA" do package "data_wrangling"
    data_sia_auxiliary = DataSiaAuxiliary(path)

    # Chama métodos da classe "DataSiaAuxiliary" do módulo "prepare_SIA" referentes ao sub-banco de dados sia_pa
    df_CADGERBR = data_sia_auxiliary.get_CADGERBR_treated()
    df_CADGERBR.to_sql('pacoduni', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TABUF = data_sia_auxiliary.get_TABUF_treated()
    df_TABUF.to_sql('ufcod', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_RSAUDE = data_sia_auxiliary.get_RSAUDE_treated()
    df_RSAUDE.to_sql('rsaude', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_CADMUN = data_sia_auxiliary.get_CADMUN_treated()
    df_CADMUN.to_sql('pagestao', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TP_GESTAO = data_sia_auxiliary.get_TP_GESTAO_treated()
    df_TP_GESTAO.to_sql('pacondic', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    # Mesmo objeto pandas DataFrame da tabela "pagestao"
    df_CADMUN.to_sql('paufmun', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_REGRA_C = data_sia_auxiliary.get_REGRA_C_treated()
    df_REGRA_C.to_sql('paregct', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TP_ESTAB = data_sia_auxiliary.get_TP_ESTAB_treated()
    df_TP_ESTAB.to_sql('patpups', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_ESFERA = data_sia_auxiliary.get_ESFERA_treated()
    df_ESFERA.to_sql('patippre', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TB_SIGTAP = data_sia_auxiliary.get_TB_SIGTAP_treated()
    df_TB_SIGTAP.to_sql('paproc', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_FINANC = data_sia_auxiliary.get_FINANC_treated()
    df_FINANC.to_sql('patpfin', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_COMPLEX = data_sia_auxiliary.get_COMPLEX_treated()
    df_COMPLEX.to_sql('panivcpl', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_DOCORIG = data_sia_auxiliary.get_DOCORIG_treated()
    df_DOCORIG.to_sql('padocorig', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_CBO = data_sia_auxiliary.get_CBO_treated()
    df_CBO.to_sql('pacbocod', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_MOTSAIPE = data_sia_auxiliary.get_MOTSAIPE_treated()
    df_MOTSAIPE.to_sql('pamotsai', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_S_CID = data_sia_auxiliary.get_S_CID_treated()
    df_S_CID.to_sql('pacidpri', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    # Mesmo objeto pandas DataFrame da tabela "pacidpri"
    df_S_CID.to_sql('pacidsec', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    # Mesmo objeto pandas DataFrame da tabela "pacidpri"
    df_S_CID.to_sql('pacidcas', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_CARAT_AT = data_sia_auxiliary.get_CARAT_AT_treated()
    df_CARAT_AT.to_sql('pacatend', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_FL_IDADE = data_sia_auxiliary.get_FL_IDADE_treated()
    df_FL_IDADE.to_sql('paflidade', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_SEXO = data_sia_auxiliary.get_SEXO_treated()
    df_SEXO.to_sql('pasexo', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_RACA_COR = data_sia_auxiliary.get_RACA_COR_treated()
    df_RACA_COR.to_sql('paracacor', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    # Mesmo objeto pandas DataFrame da tabela "pagestao"
    df_CADMUN.to_sql('pamunpcn', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_INDICA = data_sia_auxiliary.get_INDICA_treated()
    df_INDICA.to_sql('paindica', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_CODOCO = data_sia_auxiliary.get_CODOCO_treated()
    df_CODOCO.to_sql('pacodoco', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_etnia = data_sia_auxiliary.get_etnia_treated()
    df_etnia.to_sql('paetnia', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    # Mesmo objeto pandas DataFrame da tabela "pacbocod"
    df_CBO.to_sql('pasrcc', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_INE_EQUIPE_BR = data_sia_auxiliary.get_INE_EQUIPE_BR_treated()
    df_INE_EQUIPE_BR.to_sql('paine', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_natjur = data_sia_auxiliary.get_natjur_treated()
    df_natjur.to_sql('panatjur', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TB_GRUPO = data_sia_auxiliary.get_TB_GRUPO_treated()
    df_TB_GRUPO.to_sql('grupo', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TB_SUBGR = data_sia_auxiliary.get_TB_SUBGR_treated()
    df_TB_SUBGR.to_sql('subgrupo', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)

    df_TB_FORMA = data_sia_auxiliary.get_TB_FORMA_treated()
    df_TB_FORMA.to_sql('forma', con=device, schema=child_db, if_exists=label1, index=False, index_label=label2)


###########################################################################################################################
#    copy_expert+pandas copy_expert+pandas copy_expert+pandas copy_expert+pandas copy_expert+pandas copy_expert+pandas    #
###########################################################################################################################
###########################################################################################################################
#    MAIN TABLE * MAIN TABLE * MAIN TABLE * MAIN TABLE * MAIN TABLE * MAIN TABLE * MAIN TABLE * MAIN TABLE * MAIN TABLE   #
###########################################################################################################################
# Função que utiliza "copy_expert" para a inserção de dados principais e "pandas.to_sql" para a inserção
# dos respectivos metadados no banco de dados "child_db"
def insert_into_main_table_and_arquivos(file_name, directory, date_ftp, device, child_db, connection_data):
    start = time.time()
    counting_rows = pd.read_sql(f'''SELECT COUNT('NOME') FROM {child_db}.arquivos''', con=device)
    qtd_files_pg = counting_rows.iloc[0]['count']
    print(f'A quantidade de arquivos principais de dados do {child_db} já carregada no {connection_data[0]}/PostgreSQL é {qtd_files_pg}.')

    # Tratamento de dados principais do sia_xx
    base = file_name[0:2]
    state = file_name[2:4]
    year = file_name[4:6]
    month = file_name[6:8]
    # No mês de dez/2011 e em todos os meses a partir de jan/2013 os arquivos "dbc" PAXXaamm se dividem em...
    # dois por mês tendo no final "appended" a string "a" ou "b"
    if ((state + year + month == 'SP1112') or ((state == 'SP') and (int(year) >= 13))):
        month = file_name[6:9]
    main_table = base.lower() + 'br'

    print(f'\nIniciando a lida com o arquivo {base}{state}{year}{month}...')
    # Cria uma instância da classe "DataSiaMain" do módulo "prepare_SIA" do package "data_wrangling"
    data_sia_main = DataSiaMain(base, state, year, month)
    # Chama método da classe "DataSiaMain" do módulo "prepare_SIA" referentes ao sub-banco de dados sia_xx
    df = data_sia_main.get_SIAXXaamm_treated()

    # Inserção das colunas UF_XX, ANO_XX e MES_XX no objeto pandas DataFrame "df"
    df.insert(0, 'UF_' + base, [state]*df.shape[0])
    df.insert(1, 'ANO_' + base, [int('20' + year)]*df.shape[0])
    # No mês de dez/2011 e em todos os meses a partir de jan/2013 os arquivos "dbc" PAXXaamm se dividem em...
    # dois por mês tendo no final "appended" a string "a" ou "b"
    if ((state + year + month == 'SP1112') or ((state == 'SP') and (int(year) >= 13))):
        df.insert(2, 'MES_PA', [month[0:2]]*df.shape[0])
    else:
        df.insert(2, 'MES_PA', [month]*df.shape[0])

    # Criação de arquivo "csv" contendo os dados do arquivo principal de dados do sia_xx armazenado no objeto...
    # pandas DataFrame "df"
    df.to_csv(base + state + year + month + '.csv', sep=',', header=False, index=False, escapechar=' ')
    # Leitura do arquivo "csv" contendo os dados do arquivo principal de dados do sia_xx
    f = open(base + state + year + month + '.csv', 'r')
    # Conecta ao banco de dados mãe "connection_data[0]" do SGBD PostgreSQL usando o módulo python "psycopg2"
    conn = psycopg2.connect(dbname=connection_data[0],
                            host=connection_data[1],
                            port=connection_data[2],
                            user=connection_data[3],
                            password=connection_data[4])
    # Criação de um cursor da conexão tipo "psycopg2" referenciado à variável "cursor"
    cursor = conn.cursor()
    try:
        # Faz a inserção dos dados armazenados em "f" na tabela "main_table" do banco de dados "child_db"...
        # usando o método "copy_expert" do "psycopg2"
        cursor.copy_expert(f'''COPY {child_db}.{main_table} FROM STDIN WITH CSV DELIMITER AS ',';''', f)
    except:
        print(f'Tentando a inserção do arquivo {base}{state}{year}{month} por método alternativo (pandas)...')
        df.to_sql(main_table, con=device, schema=child_db, if_exists='append', index=False)
    else:
        conn.commit()
    # Encerra o cursor
    cursor.close()
    # Encerra a conexão
    conn.close()
    # Encerra o file handler
    f.close()
    # Remoção do arquivo "csv"
    os.remove(base + state + year + month + '.csv')
    print(f'Terminou de inserir os dados do arquivo {base}{state}{year}{month} na tabela {main_table} do banco de dados {child_db}.')

    # Cria um objeto pandas DataFrame com apenas uma linha de dados, a qual contém informações sobre o...
    # arquivo de dados principal carregado
    file_data = pd.DataFrame(data=[[file_name, directory, date_ftp, datetime.today(), int(df.shape[0])]],
                             columns= ['NOME', 'DIRETORIO', 'DATA_INSERCAO_FTP', 'DATA_HORA_CARGA', 'QTD_REGISTROS'],
                             index=None
                             )
    # Inserção de informações do arquivo principal de dados no banco de dados "child_db"
    file_data.to_sql('arquivos', con=device, schema=child_db, if_exists='append', index=False)
    print(f'Terminou de inserir os metadados do arquivo {base}{state}{year}{month} na tabela arquivos do banco de dados {child_db}.')
    end = time.time()
    print(f'Demorou {round((end - start)/60, 1)} minutos para essas duas inserções no {connection_data[0]}/PostgreSQL!')



"""
Cria o diretório datasus_content no home directory do computador de execução deste módulo.

Código elaborado por Flávio Coelho (https://github.com/fccoelho/PySUS).
"""

import os
from pathlib import Path


if not os.path.exists(os.path.join(Path.home(), 'datasus_content')):
    os.mkdir(os.path.join(Path.home(), 'datasus_content'))
CACHEPATH = os.path.join(Path.home(), 'datasus_content')


def cache_contents():
    """
    List the files currently cached in ~/datasus_content
    :return:
    """
    cached_data = os.listdir(CACHEPATH)
    return [os.path.join(CACHEPATH, f) for f in cached_data]



import os
import setuptools

lg = """
The PegaSUS project is divided into n modules. One of the implemented modules,
called dados, allows to create a schema of a PostgreSQL database, previously
created in the user computer, and to insert data that belongs to the following
information health systems of the Brazilian Public Health System (SUS): CNES,
SIH, SIA, SIM, SINASC or SINAN (more to come). The dados module also also allows
to directly read main data files (in DBC format) and auxiliary data files (in
DBF or CNV format) as pandas DataFrame objects. This second option is recommended
in the case of more punctual consultation.

To use the dados module by means of a PostgreSQL database it is necessary to
install the PostgreSQL DBMS (https://www.postgresql.org/download/) and it is
recommended to have a database management software such as pgAdmin (https://www.pgadmin.org/download/)
or DBeaver Community (https://dbeaver.io/), to proper visualize the structured
data.

To use the dados module on Linux or MacOS platforms it is necessary to install
the libffi dependency:
1) Linux: `$ sudo apt install libffi-dev`
2) MacOS: `$ brew install libffi`

For use on Windows platforms it is necessary to install the executable program
(*.exe*) dbf2dbc, which is part of the Datasus Tabwin program. To accomplish this,
download the "Tab415.zip" folder by selecting the eponymous link contained in the
first line of the "Name" column of the table that appears in http://www2.datasus.gov.br/DATASUS/index.php?area=060805&item=3.
Unpack the "Tab415.zip" folder, copy the executable "dbf2dbc.exe" and the file
"IMPBORL.DLL" (the unpacked folder "Tab415" and its *.zip* can be deleted) to a
newly created folder named "dbf2dbc" in the root directory of the user's computer,
which generally has the following form on Windows platforms:

```C:\\```

Finally, the data module is installed using the command(on Linux platforms):

```$ sudo pip install pegasus-dados```
"""

if os.name == 'posix':

    setuptools.setup(
        name="pegasus-dados",
        version="0.2.0",
        packages=setuptools.find_packages(),
        package_data={
            '': ['*.xlsx', '*.c', '*.h', '*.o', '*.so', '*.md']
        },
        author="SecexSaudeTCU",
        author_email="ericlb@tcu.gov.br",
        description="Plataforma Eletrônica de Governança e Accountability do SUS - Módulo dados",
        long_description=lg,
        long_description_content_type="text/markdown",
        url="https://github.com/SecexSaudeTCU/pegasus/tree/master/pegasus/dados",
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License"
        ],
        python_requires='>=3.6',
        setup_requires=['cffi>=1.0.0'],
        cffi_modules=["pegasus/dados/transform/extract/_build_readdbc.py:ffibuilder"],
        install_requires=['numpy', 'pandas', 'xlrd', 'dbfread', 'cffi>=1.0.0',
                          'pyarrow', 'fastparquet', 'psycopg2', 'SQLAlchemy']
    )

elif os.name == 'nt':

    setuptools.setup(
        name="pegasus-dados",
        version="0.2.0",
        packages=setuptools.find_packages(),
        package_data={
            '': ['*.xlsx', '*.c', '*.h', '*.o', '*.so', '*.md']
        },
        author="SecexSaudeTCU",
        author_email="ericlb@tcu.gov.br",
        description="Plataforma Eletrônica de Governança e Accountability do SUS - Módulo dados",
        long_description=lg,
        long_description_content_type="text/markdown",
        url="https://github.com/SecexSaudeTCU/pegasus/tree/master/pegasus/dados",
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License"
        ],
        python_requires='>=3.6',
        install_requires=['numpy', 'pandas', 'xlrd', 'dbfread', 'pyarrow', 'fastparquet', 'psycopg2', 'SQLAlchemy']
    )
